;;;; JUI - sudoku (language demo), v1
;;;  (c) MM, 2012

;;;; Zadanie: riesenie sudoku
;;;
;;;  Hlavolam pozostava z mriezky vytvorenej deviatimi riadkami, deviatimi
;;;  stlpcami a deviatimi 3x3 stvorcami. Spolu obsahuje 81 pozicii, pricom
;;;  niektore pozicie su obsadené, zatial co ine su volne. Cielom je na 
;;;  kazdu volnu poziciu priradit jednu cislicu z {1,2,3,4,5,6,7,8, } tak,
;;;  aby v ziadnom riadku, stlpci ani stvorci neboli dve rovnake cislice.

(defparameter sud1 '(1 * * 8 3 * * * 2
		     5 7 * * * 1 * * *
		     * * * 5 * 9 * 6 4
		     7 * 4 * * 8 5 9 *
		     * * 3 * 1 * 4 * *
		     * 5 1 4 * * 3 * 6
		     3 6 * 7 * 4 * * *
		     * * * 6 * * * 7 9
		     8 * * * 5 2 * * 3) 
  "lahka (Gould, W.: Pravda Sudoku kniha 1. Bratislava: Denník Pravda 2005)")

(defparameter sud2 '(* * 7 3 * * 2 * *
		     3 * * * * * * * 1
		     8 * * 6 2 * * * *
		     * 7 3 4 * * * * 5
		     * * * * * * * * *
		     5 * * * * 8 4 9 *
		     * * * * 6 7 * * 4
		     2 * * * * * * * 6
		     * * 9 * * 4 3 * *) 
  "zakerna (Gould, W.: Pravda Sudoku kniha 1. Bratislava: Denník Pravda 2005)")

(defparameter sud3  '(1 * * * * 7 * 9 *
		      * 3 * * 2 * * * 8
		      * * 9 6 * * 5 * *
		      * * 5 3 * * 9 * *
		      * 1 * * 8 * * * 2
		      6 * * * * 4 * * *
		      3 * * * * * * 1 *
		      * 4 * * * * * * 7
		      * * 7 * * * 3 * *) 
  "AI-escargot (Inkala, A.: AI Sudoku Top 10, 2006, http://www.aisudoku.com/index_en.html)")


;;;; Verzia 1 *******************************************************
;;;
;;;  Navrh datovych struktur, pomocne pristupove funkcie 

(defun vytvor-mriezku (hodnoty)
  (let (mriezka)
    (dotimes (i 81)
      (push (cons (list (floor (/ i 9))              ; riadok
			(mod i 9)                    ; stlpec
			(+ (* 3 (floor (/ i 27))) 
			   (floor (/ (mod i 9) 3)))) ; stvorec
		  (if (eql (nth i hodnoty) '*) 
		      (list 1 2 3 4 5 6 7 8 9)       ; vsetky moznosti
		      (list (nth i hodnoty))))       ; iba dana moznost
	    mriezka))
    mriezka))

(defun poz-suradnice (x)
  (car x))

(defun poz-hodnoty (x)
  (cdr x))

#|
  (vytvor-mriezku sud1)

  (car (vytvor-mriezku sud1))
  (poz-suradnice (car (vytvor-mriezku sud1)))
  (poz-hodnoty (car (vytvor-mriezku sud1)))
|#


;;;; Verzia 2 *******************************************************
;;;
;;;  Vizualizacia dat

(defun poc-kombinacii (mriezka)
  (reduce #'(lambda (x y) (* x (length (poz-hodnoty y)))) 
	  mriezka :initial-value 1))

(defun tlac-mriezku (mriezka)
  (format t "~{~%[~{~a~}][~{~a~}][~{~a~}][~{~a~}][~{~a~}][~{~a~}][~{~a~}][~{~a~}][~{~a~}]~}~%" 
	  (mapcar #'(lambda (x) 
		      (mapcar #'(lambda (y) 
				  (if (member y (poz-hodnoty x)) y #\space)) 
			      (list 1 2 3 4 5 6 7 8 9))) 
		  (sort (copy-list mriezka) 
			#'(lambda (x y) (or (< (car x) (car y))
					    (and (= (car x) (car y))
						 (< (cadr x) (cadr y)))))
			:key #'poz-suradnice) )))

#|
  (poc-kombinacii (vytvor-mriezku sud1))

  (tlac-mriezku (vytvor-mriezku sud1))
|#


;;;; Verzia 3 *******************************************************
;;;
;;;  API pre zmenu dat

(defun redukuj-poziciu (suradnice hodnoty mriezka)
  (let ((pozicia (member suradnice mriezka :key #'poz-suradnice :test #'equal)))
    (setf (car pozicia)
	  (cons suradnice 
		(set-difference (poz-hodnoty (car pozicia)) hodnoty))))
  mriezka)

#|
  (tlac-mriezku (vytvor-mriezku sud1))
  (tlac-mriezku (redukuj-poziciu '(0 1 0) '(3 5) (vytvor-mriezku sud1)))
|#


;;;; Verzia 4 *******************************************************
;;;
;;;  API pre vyber dat

(defun eqlpos (vzor suradnice)
  (and (or (eql '* (car vzor)) (eql (car vzor) (car suradnice)))
       (or (eql '* (cadr vzor)) (eql (cadr vzor) (cadr suradnice)))
       (or (eql '* (caddr vzor)) (eql (caddr vzor) (caddr suradnice)))))

(defun vyber-riadok (x mriezka)
  (remove-if-not #'(lambda (pozicia) 
		     (eqlpos `(,x * *) (poz-suradnice pozicia))) 
		 mriezka))

(defun vyber-stlpec (x mriezka)
  (remove-if-not #'(lambda (pozicia) 
		     (eqlpos `(* ,x *) (poz-suradnice pozicia))) 
		 mriezka))

(defun vyber-stvorec (x mriezka)
  (remove-if-not #'(lambda (pozicia) 
		     (eqlpos `(* * ,x) (poz-suradnice pozicia))) 
		 mriezka))

(defun vyber-utvary (mriezka)
  (mapcan #'(lambda (x) (list (vyber-riadok x mriezka)
			      (vyber-stlpec x mriezka)
			      (vyber-stvorec x mriezka)))
	  (list 0 1 2 3 4 5 6 7 8)))

#|
  (eqlpos '(1 1 1) '(1 1 1))
  (eqlpos '(1 * 1) '(1 1 1))

  (tlac-mriezku (vytvor-mriezku sud1))
  (vyber-riadok 0 (vytvor-mriezku sud1))
  (vyber-stlpec 0 (vytvor-mriezku sud1))
  (vyber-stvorec 0 (vytvor-mriezku sud1))

  (length (vyber-utvary (vytvor-mriezku sud1)))
|#


;;;; Verzia 5 *******************************************************
;;;
;;;  Filtracia pozicii

(defun najdi-pozicie-s-hodnotami (hodnoty utvar &optional (vylucne nil))
  (let ((pozicie utvar))
    (dolist (x hodnoty)
      (setf pozicie (remove-if-not #'(lambda (y) (member x (poz-hodnoty y))) 
				   pozicie)))
    (if vylucne
	(remove-if-not #'(lambda (x) (= (length (poz-hodnoty x)) 
					(length hodnoty)))
		       pozicie)
	pozicie)))

#|
  (tlac-mriezku (vytvor-mriezku sud1))
  (najdi-pozicie-s-hodnotami '(3) (vyber-riadok 0 (vytvor-mriezku sud1)))
  (najdi-pozicie-s-hodnotami '(3) (vyber-riadok 0 (vytvor-mriezku sud1)) t)
  (najdi-pozicie-s-hodnotami '(3 2 1) (vyber-riadok 0 (vytvor-mriezku sud1)))
|#


;;;; Verzia 6 *******************************************************
;;;
;;;  Redukcia 1: 'naked individuals'
;;;  pre nejaku poziciu je nepripustna taka hodnota, ktora je uz umiestnena
;;;  v utvare (riadku, stlpci alebo stvorci), do ktoreho patri dana pozicia,
;;;  na nejakej inej pozicii
;;;  -> ak nejaka hodnota je v utvare umiestnena na nejakej pozicii, tak sa
;;;     odstrania vyskyty tejto hodnoty na ostatnych poziciach utvaru 

(defun r1-utvar (utvar mriezka)
  (mapc #'(lambda (x) 
	    (let ((pozicie (najdi-pozicie-s-hodnotami x utvar :vylucne)))
	      (when (= 1 (length pozicie))
		(dolist (y (set-difference utvar pozicie 
					   :key #'poz-suradnice :test #'equal))
		  (redukuj-poziciu (poz-suradnice y) x mriezka)))))
	(mapcar #'list (list 1 2 3 4 5 6 7 8 9)))
  mriezka)

(defun r1 (mriezka)
  (mapc #'(lambda (x) (r1-utvar x mriezka)) (vyber-utvary mriezka))
  mriezka)

(defun ries-r1 (mriezka)
  (do ((pred (poc-kombinacii mriezka) po) (po 0) koniec)
      (koniec mriezka)
    (r1 mriezka)
    (setf po (poc-kombinacii mriezka))
    (when (= pred po)
      (setf koniec t))))

#|
  (tlac-mriezku (vytvor-mriezku sud1))
  (tlac-mriezku (r1-utvar (vyber-riadok 0 (vytvor-mriezku sud1)) (vytvor-mriezku sud1)))

  (poc-kombinacii (vytvor-mriezku sud1))
  (poc-kombinacii (r1 (vytvor-mriezku sud1)))
  (poc-kombinacii (r1 (r1 (vytvor-mriezku sud1))))
  ...

  (poc-kombinacii (ries-r1 (vytvor-mriezku sud1)))
  (tlac-mriezku (ries-r1 (vytvor-mriezku sud1)))

  (poc-kombinacii (ries-r1 (vytvor-mriezku sud2)))
|#


;;;; Verzia 7 *******************************************************
;;;
;;;  Zovseobecnenie redukcie 1: 
;;;    1 hodnota - 'naked individuals'
;;;    2 hodnoty - 'naked pairs'
;;;    3 hodnoty - 'naked triples'
;;;    ...

(defun kombinacie (elements counter &optional (combinations '(nil)))
  (cond ((zerop counter) 
	 (remove-if-not #'(lambda (x) (apply #'< x)) combinations))
	(t (kombinacie elements 
		       (1- counter)
		       (mapcan #'(lambda (x)
				   (mapcar #'(lambda (y) (cons y x)) elements))
			       combinations)))))

(defun r1-utvar (mohutnost utvar mriezka)
  (mapc #'(lambda (x) 
	    (let ((pozicie (najdi-pozicie-s-hodnotami x utvar :vylucne)))
	      (when (= mohutnost (length pozicie))
		(dolist (y (set-difference utvar pozicie 
					   :key #'poz-suradnice :test #'equal))
		  (redukuj-poziciu (poz-suradnice y) x mriezka)))))
	(kombinacie (list 1 2 3 4 5 6 7 8 9) mohutnost))
  mriezka)

(defun r1 (mohutnost mriezka)
  (mapc #'(lambda (x) (r1-utvar mohutnost x mriezka)) (vyber-utvary mriezka))
  mriezka)

(defun ries-r1 (mohutnost mriezka)
  (do ((pred (poc-kombinacii mriezka) po) (po 0) koniec)
      (koniec mriezka)
    (r1 mohutnost mriezka)
    (setf po (poc-kombinacii mriezka))
    (when (= pred po)
      (setf koniec t))))

#|
  (kombinacie '(1 2 3 4 5) 1)
  (kombinacie '(1 2 3 4 5) 2)
  (kombinacie '(1 2 3 4 5) 3)
  ...
  (length (kombinacie '(1 2 3 4 5 6 7 8 9) 6))

  (poc-kombinacii (vytvor-mriezku sud1))
  (poc-kombinacii (r1 1 (vytvor-mriezku sud1)))
  (poc-kombinacii (r1 2 (r1 1 (vytvor-mriezku sud1))))
  (poc-kombinacii (r1 1 (r1 2 (r1 1 (vytvor-mriezku sud1)))))
  (poc-kombinacii (r1 2 (r1 1 (r1 2 (r1 1 (vytvor-mriezku sud1))))))

  (poc-kombinacii (ries-r1 1 (vytvor-mriezku sud1)))

  (poc-kombinacii (ries-r1 1 (vytvor-mriezku sud2)))
  (poc-kombinacii (r1 2 (ries-r1 1 (vytvor-mriezku sud2))))
  (poc-kombinacii (r1 3 (ries-r1 1 (vytvor-mriezku sud2))))
|#


;;;; Verzia 8 *******************************************************
;;;
;;;  Redukcia 2:
;;;  ak v nejakom utvare (riadku, stlpci alebo stvorci) je pre nejaku hodnotu
;;;  pripustna iba jedna pozicia, tak ta hodnota sa umiestni na danu poziciu
;;;  bez ohladu na to, ci pre danu poziciu su pripustne aj dalsie hodnoty
;;;  alebo nie.
;;;  -> ak pre nejaku hodnotu je v utvare pripustna iba jedna pozicia, tak 
;;;     z tej pozicie sa odstrania vsetky ostatne hodnoty
;;;
;;;  Zovseobecnenie redukcie 2: 
;;;    1 hodnota - 'hiden individuals'
;;;    2 hodnoty - 'hiden pairs'
;;;    3 hodnoty - 'hiden triples'
;;;    ...

(defun r2-utvar (mohutnost utvar mriezka)
  (mapc #'(lambda (x) 
	    (let ((pozicie (najdi-pozicie-s-hodnotami x utvar))
		  (pozicie2 
		   (reduce #'union 
			   (mapcar #'(lambda (y) 
				       (najdi-pozicie-s-hodnotami (list y) utvar)) 
				   x))))
	      (when (= mohutnost (length pozicie) (length pozicie2))
		(dolist (y pozicie)
		  (redukuj-poziciu (poz-suradnice y) 
				   (set-difference (poz-hodnoty y) x)
				   mriezka)))))
	(kombinacie (list 1 2 3 4 5 6 7 8 9) mohutnost))
  mriezka)

(defun r2 (mohutnost mriezka)
  (mapc #'(lambda (x) (r2-utvar mohutnost x mriezka)) (vyber-utvary mriezka))
  mriezka)

#|
  (tlac-mriezku (ries-r1 1 (vytvor-mriezku sud2)))
  (tlac-mriezku (r2-utvar 1 (vyber-stvorec 6 (ries-r1 1 (vytvor-mriezku sud2))) (ries-r1 1 (vytvor-mriezku sud2))))
  (tlac-mriezku (r2-utvar 2 (vyber-stlpec 5 (ries-r1 1 (vytvor-mriezku sud2))) (ries-r1 1 (vytvor-mriezku sud2))))

  (poc-kombinacii (ries-r1 1 (vytvor-mriezku sud2)))
  (poc-kombinacii (r2 1 (ries-r1 1 (vytvor-mriezku sud2))))
  (poc-kombinacii (r2 2 (ries-r1 1 (vytvor-mriezku sud2))))
  (poc-kombinacii (r2 2 (r2 1 (ries-r1 1 (vytvor-mriezku sud2)))))
|#


;;;; Verzia 9 *******************************************************
;;;
;;;  Zostavenie redukcneho algoritmu

(defun redukuj (mriezka 
		&optional (metody '((r1 . 1) (r2 . 1) (r1 . 2) 
				    (r2 . 2) (r1 . 3) (r2 . 3))))
  (do ((pred (poc-kombinacii mriezka) po) (po 0) koniec)
      (koniec mriezka)
    (dolist (x (sort (copy-list metody) 
		     #'< :key 
		     #'(lambda (y) (declare (ignore y)) (random 1.0))))
      ;; (format t "~a-~a " (car x) (cdr x))
      (funcall (symbol-function (car x)) (cdr x) mriezka))
    (setf po (poc-kombinacii mriezka))
    (when (= pred po)
      (setf koniec t))))

#|
  ;; zapnut formatovany vypis
  (poc-kombinacii (redukuj (vytvor-mriezku sud2) '((r1 . 1)(r1 . 2)(r1 . 3))))

  (poc-kombinacii (redukuj (vytvor-mriezku sud2) '((r2 . 1)(r2 . 2)(r2 . 3))))

  (poc-kombinacii (redukuj (vytvor-mriezku sud2) '((r1 . 1)(r2 . 1))))
  (poc-kombinacii (redukuj (vytvor-mriezku sud2) '((r1 . 1)(r2 . 1)(r1 . 2))))
  (poc-kombinacii (redukuj (vytvor-mriezku sud2) '((r1 . 1)(r2 . 1)(r2 . 2))))

  (poc-kombinacii (redukuj (vytvor-mriezku sud2)))

  (poc-kombinacii (redukuj (vytvor-mriezku sud3)))
  ;; vypnut formatovany vypis
|#


;;;; Verzia 10 ******************************************************
;;;
;;;  Pridanie redukcie pokus-omyl

(defun dosad (mriezka)
  (let* ((nevyriesene-pozicie
	  (remove-if #'(lambda (x) (= 1 (length x))) 
		     mriezka 
		     :key #'poz-hodnoty))
	 (pozicia (nth (random (length nevyriesene-pozicie)) 
		       nevyriesene-pozicie)))
    ;; (format t "D-~a " (length (poz-hodnoty pozicia)))
    (mapcar #'(lambda (x)
		(let ((nova-mriezka (copy-list mriezka)))
		  (setf (car (member (poz-suradnice pozicia) 
				     nova-mriezka 
				     :key #'poz-suradnice :test #'equal))
			(cons (poz-suradnice pozicia) (list x)))
		  nova-mriezka))
	    (poz-hodnoty pozicia))))

(defun sudoku (mriezka)
  (redukuj mriezka '((r1 . 1) (r1 . 2) (r2 . 1) (r2 . 2)))
  (cond ((< 1 (poc-kombinacii mriezka))
	 (car (remove-if #'(lambda (x) (/= 1 (poc-kombinacii x))) 
			 (mapcar #'sudoku (dosad mriezka)))))
	(t mriezka)))

#|
  ;;; odkomentovat formatovy vypis
 (poc-kombinacii (sudoku (vytvor-mriezku sud2)))

 (poc-kombinacii (sudoku (vytvor-mriezku sud3)))
 (tlac-mriezku (sudoku (vytvor-mriezku sud3)))
    ;;; ak ano, tak v niektorom utvare nejaka hodnota chyba alebo je viackrat
    ;;; inokedy tlaci nil
|#


;;;; Verzia 11 ******************************************************
;;;
;;;  Kontrola nevalidnych mriezok

(defun valid (mriezka)
  (every #'(lambda (utvar)
	     (every #'(lambda (hodnota) 
		      (some #'(lambda (pozicia) 
				(member hodnota (poz-hodnoty pozicia))) 
		       utvar))
		    (list 1 2 3 4 5 6 7 8 9)))
	 (vyber-utvary mriezka)))

#|
  (defparameter qq 0)
  (setf qq (sudoku (vytvor-mriezku sud3)))
     ;; tak aby tam bola nejaka mriezka a nie nil
  (valid qq)
  (tlac-mriezku qq)
|#


;;;; Verzia 12 ******************************************************
;;;
;;;  Oprava - vyhadzovanie nevalidnych mriezok

(defun sudoku (mriezka)
  (redukuj mriezka '((r1 . 1) (r2 . 1)))
  (cond ((not (valid mriezka)) nil)
	((< 1 (poc-kombinacii mriezka))
	 (car (remove-if #'(lambda (x) (/= 1 (poc-kombinacii x))) 
			 (remove nil (mapcar #'sudoku (dosad mriezka))))))
	(t mriezka)))

#|
  (setf qq (sudoku (vytvor-mriezku sud3)))
  (valid qq)
  (tlac-mriezku qq)
|#
